/**
 * @file buddy.c
 *
 * does everything with SkypeAPI buddies: status, info, etc.
 *
 * Copyright (C) 2007, Willem van Engen <dev-pidgin@willem.engen.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <string.h>
#include <glib/gprintf.h>

#include <debug.h>
#include <blist.h>
#include <util.h>
#include <server.h>

#include "skypeapi.h"
#include "buddy.h"
#include "conn.h"

// callbacks
static gboolean skypeapi_buddy_get_avatar_cb_rmfile(gpointer p);
static gboolean skypeapi_buddy_cb_addbuddies(gpointer _gc, char *data, gpointer p);
static gboolean skypeapi_buddy_pcb_get_info_2(gpointer _gc, char *data, gpointer p);

// local functions
static gboolean skypeapi_buddy_add_byname_internal(PurpleConnection *gc, const char *who);
static gboolean skypeapi_buddy_rem_byname_internal(PurpleConnection *gc, const char *who);


// We need to store data here to
struct skypeapi_buddy_data {
	GSList *buddies;	//< List of buddies, containing struct skypeapi_buddy_buddy_data
};
 
 // What we need to store for a single buddy
struct skypeapi_buddy_buddy_data {
	char *name;
	char *status_text;
};

static void skypeapi_buddy_buddy_free(struct skypeapi_buddy_buddy_data *bud)
{
	if (bud->status_text) g_free(bud->status_text);
	g_free(bud->name);
	g_free(bud);
}

// return data for a buddy identified by his name
static struct skypeapi_buddy_buddy_data
	*skypeapi_buddy_get_byname(struct skypeapi_buddy_data *bd, const char *who)
{
	GSList *cur = bd->buddies;

	if (!cur) return NULL;
	
	do {
		struct skypeapi_buddy_buddy_data *bud =
			(struct skypeapi_buddy_buddy_data*)cur->data;
		if (strcmp(who, bud->name)==0) {
			return bud;
		}		
	} while( (cur=g_slist_next(cur)) );
	
	// found none
	return NULL;
}

// called when skype emits a "USER" message (without message id)
static gboolean skypeapi_buddy_cb_user(gpointer _gc, char *cmd, gpointer data)
{
	PurpleConnection *gc = (PurpleConnection *)_gc;
	PurpleAccount *acc = purple_connection_get_account(gc);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	struct skypeapi_buddy_data *bd = (struct skypeapi_buddy_data*)sa->d_buddy;
	struct skypeapi_buddy_buddy_data *bud;
	char *who, *prop, *c;
	PurpleBuddy *pbuddy;

	// format: (USER) <id> property <value>	
	who = strtok(cmd, skypeapi_conn_split_word);
	prop = strtok(NULL, skypeapi_conn_split_word);

	// Now handle user updates
	if (!g_ascii_strcasecmp(prop, "FULLNAME")) {
		c = strtok(NULL, skypeapi_conn_split_eol);
		serv_got_alias(gc, who, c);
	} else if (!g_ascii_strcasecmp(prop, "ONLINESTATUS")) {
		c = strtok(NULL, skypeapi_conn_split_word);
		purple_debug_info("skypeapi", "buddy: got status for '%s': '%s'\n", who, c);
		purple_prpl_got_user_status(acc, who, c, NULL);
	} else if (!g_ascii_strcasecmp(prop, "MOOD_TEXT")) {
		c = strtok(NULL, skypeapi_conn_split_eol);
		// store locally, notify purple to get it later
		if ( (bud=skypeapi_buddy_get_byname(bd, who)) ) {
			bud->status_text = g_strdup(c);
		}
		// TODO update ui, how?
		//purple_prpl_got_user_status(acc, who, 
	} else if (!g_ascii_strcasecmp(prop, "BUDDYSTATUS")) {
		c = strtok(NULL, skypeapi_conn_split_word);
		switch( atoi(c) ) {
		case 1: // deleted from contact list, make sure it's gone here too
			skypeapi_buddy_rem_byname_internal(gc, who);
			// Remove from purple's list
			if ( (pbuddy = purple_find_buddy(acc, who)) ) {
				purple_blist_remove_buddy(pbuddy);
				purple_debug_info("skypeapi", "buddy: '%s' deleted from contact list\n", who);
			}
			break;
		case 2: // authorization requested
		case 3: // accepted by remote side
			if (skypeapi_buddy_add_byname_internal(gc, who))
				purple_debug_info("skypeapi", "buddy: '%s' added to contact list\n", who);
			break;
		}
	} else if (!g_ascii_strcasecmp(prop, "AVATAR")) {
		c = strtok(NULL, skypeapi_conn_split_word);
		if (atoi(c)!=1)
			purple_debug_warning("skypeapi", "Unrecognized avatar number %d\n", atoi(c));
		c = strtok(NULL, skypeapi_conn_split_word);
		// TODO: get avatar from file c and put it into my buddy struct
	}
	
	// Keep me notified!
	return TRUE;
}

static void skypeapi_buddy_get_avatar(PurpleConnection *gc, const char *who)
{
	PurpleAccount *acc = purple_connection_get_account(gc);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	char *tmpfile;
	
	return; // BUG avatars don't seem to be working on Linux :(s

	// Get avatar into temporary file
	tmpfile = g_strdup_printf("%s" G_DIR_SEPARATOR_S "skypeavatarXXXXX", g_get_tmp_dir());
	skypeapi_conn_send(&sa->conn, "GET USER %s AVATAR 1 %s\n", who, tmpfile);
	
	// But delete it after some seconds in case we get no reply from Skype
	// this will free the string too
	g_timeout_add_seconds(5, skypeapi_buddy_get_avatar_cb_rmfile, (gpointer)tmpfile);
}
static gboolean skypeapi_buddy_get_avatar_cb_rmfile(gpointer p)
{
	g_unlink((char*)p);
	g_free(p);
	return FALSE; // don't call me again
}

// adds a buddy by name, duplicates are ignored
// only added to purple & internal list, not to Skype
static gboolean skypeapi_buddy_add_byname_internal(PurpleConnection *gc, const char *who)
{
	PurpleAccount *acc = purple_connection_get_account(gc);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	struct skypeapi_buddy_data *bd = (struct skypeapi_buddy_data*)sa->d_buddy;
	struct skypeapi_buddy_buddy_data *bud;
	PurpleBuddy *pbuddy;

	// If it's me, just skip it.
	if (strcmp(who, purple_account_get_username(acc))==0)
		return FALSE;

	// Create buddy in protocol plugin
	bud = skypeapi_buddy_get_byname(bd, who);
	if (!bud) {
		bud = g_new0(struct skypeapi_buddy_buddy_data, 1);
		g_return_if_fail(bud!=NULL);
		bud->name = g_strdup(who);
		bd->buddies = g_slist_prepend(bd->buddies, bud);
		// Now get relevant fields for user
		skypeapi_conn_send(&sa->conn, "GET USER %s FULLNAME\n", who);
		skypeapi_conn_send(&sa->conn, "GET USER %s MOOD_TEXT\n", who);
		skypeapi_conn_send(&sa->conn, "GET USER %s ONLINESTATUS\n", who);
		skypeapi_buddy_get_avatar(gc, who);
	}
	
	// Make sure it exists in purple's list
	pbuddy = purple_find_buddy(acc, who);
	if (!pbuddy) {
		PurpleGroup *pgroup = purple_find_group(_("Buddies"));
		purple_debug_info("skypeapi", "buddy: adding %s\n", who);
		if (!pgroup) {
			pgroup = purple_group_new(_("Buddies"));
			purple_blist_add_group(pgroup, NULL);
		}
		pbuddy = purple_buddy_new(acc, who, NULL);
		purple_blist_add_buddy(pbuddy, NULL, pgroup, NULL); 			
	}	

	return pbuddy || bud;
}

// this only deletes it from internal list, not purple's.
// doing so would crash purple.
static gboolean skypeapi_buddy_rem_byname_internal(PurpleConnection *gc, const char *who)
{
	PurpleAccount *acc = purple_connection_get_account(gc);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	struct skypeapi_buddy_data *bd = (struct skypeapi_buddy_data*)sa->d_buddy;
	struct skypeapi_buddy_buddy_data *bud;

	// Remove buddy internally
	bud = skypeapi_buddy_get_byname(bd, who);
	if (bud) {
		bd->buddies = g_slist_remove(bd->buddies, bud);
		skypeapi_buddy_buddy_free(bud);
	}

	return bud!=NULL;
}

void skypeapi_buddy_init(PurpleConnection *gc)
{
	PurpleAccount *acc = purple_connection_get_account(gc);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	struct skypeapi_buddy_data *bd;
	GList *l;

	purple_debug_info("skypeapi", "buddy: initializing\n");
	
	// Allocate our own space
	sa->d_buddy = g_new0(struct skypeapi_buddy_data, 1);
	g_return_if_fail(sa->d_buddy!=NULL);
	bd = (struct skypeapi_buddy_data*)sa->d_buddy;
	bd->buddies = NULL;

	// Register API commands
	skypeapi_conn_add_callback(&sa->conn, "USER", skypeapi_buddy_cb_user, NULL);

	// Get status of all buddies on purple's list, so old ones are removed
	l = (GList*)purple_find_buddies(acc, NULL);
	do {
		skypeapi_conn_send(&sa->conn, "GET USER %s BUDDYSTATUS\n", purple_buddy_get_name(l->data) );
	} while ( (l=l->next) );

	// Get all buddies on Skype's list
	skypeapi_conn_send_callback(&sa->conn,
		skypeapi_buddy_cb_addbuddies, NULL, "SEARCH FRIENDS\n");
}
static gboolean skypeapi_buddy_cb_addbuddies(gpointer _gc, char *data, gpointer p)
{
	PurpleConnection *gc = (PurpleConnection *)_gc;
	struct skypeapi_purpleconn *sa = gc->proto_data;
	char *c;
	// skip command (which should be "USER")
	c = strtok(data, skypeapi_conn_split_word);
	// add every single friend to contact list
	while ( (c = strtok(NULL, skypeapi_conn_split_seq)) ) {
		skypeapi_conn_send(&sa->conn, "GET USER %s BUDDYSTATUS\n", c);
	}
	return FALSE;
}

void skypeapi_buddy_cleanup(PurpleConnection *gc)
{
	struct skypeapi_purpleconn *sa;
	struct skypeapi_buddy_data *bd;

	// might just happen on very early logout
	if (!gc) return;

	sa = gc->proto_data;
	if (!sa) return;
	bd = (struct skypeapi_buddy_data*)sa->d_buddy;
	if (!bd) return;

	purple_debug_info("skypeapi", "buddy: cleaning up\n");

	if (bd->buddies) g_slist_foreach(bd->buddies, (GFunc)skypeapi_buddy_buddy_free, NULL);
	g_free(bd);
}


/*
 * Purple callbacks
 */

void skypeapi_buddy_pcb_add(PurpleConnection *gc, PurpleBuddy *buddy, PurpleGroup *group)
{
	PurpleAccount *acc = purple_connection_get_account(gc);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	struct skypeapi_buddy_data *bd = (struct skypeapi_buddy_data*)sa->d_buddy;
	const char *who = purple_buddy_get_name(buddy);

	// don't add someone twice
	if (skypeapi_buddy_get_byname(bd, who))
		return;

	// add it to Skype
	// TODO append a message to command for message to be displayed on remote end
	skypeapi_conn_send(&sa->conn, "SET USER %s BUDDYSTATUS 2\n", who);
}

void skypeapi_buddy_pcb_rem(PurpleConnection *gc, PurpleBuddy *buddy, PurpleGroup *group)
{
	PurpleAccount *acc = purple_connection_get_account(gc);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	const char *who = purple_buddy_get_name(buddy);

	// TODO how about buddies in multiple groups in purple???
	skypeapi_conn_send(&sa->conn, "SET USER %s BUDDYSTATUS 1\n", who);
}

char *skypeapi_buddy_pcb_status_text(PurpleBuddy *buddy)
{
	PurpleConnection *gc = purple_account_get_connection(buddy->account);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	struct skypeapi_buddy_data *bd = (struct skypeapi_buddy_data*)sa->d_buddy;
	struct skypeapi_buddy_buddy_data *bud;
	
	purple_debug_info("skypeapi", "Status text requested\n");

	bud=skypeapi_buddy_get_byname(bd, buddy->name);
	if (!bud) return NULL;

	purple_debug_info("skypeapi", "Status text requested: %s\n", bud->status_text);

	return bud->status_text;
}

// used only in skypeapi_buddy_cb_get_info and it's callback
struct skypeapi_buddy_userdata_item
{
	PurpleNotifyUserInfo *user_info;
	const char *header;
	const char *title;
	int skipwords;
	gboolean final;
};

void skypeapi_buddy_pcb_get_info(PurpleConnection *gc, const char *who)
{
	struct skypeapi_purpleconn *sa = gc->proto_data;
	struct skypeapi_buddy_data *bd = (struct skypeapi_buddy_data*)sa->d_buddy;
	struct skypeapi_buddy_userdata_item *bui;
	PurpleNotifyUserInfo *user_info;

	purple_debug_info("skypeapi", "get info\n");
	
	user_info = purple_notify_user_info_new();

#define SKYPEAPI_GET_INFO_ITEM(HEADER, TITLE, ITEM, SKIPWORDS, FINAL) \
	{ \
		bui = g_new0(struct skypeapi_buddy_userdata_item, 1); \
		g_return_if_fail(bui!=NULL); \
		bui->user_info = user_info; \
		bui->header = (HEADER); \
		bui->title = (TITLE); \
		bui->skipwords = (SKIPWORDS); \
		bui->final = (FINAL); \
		skypeapi_conn_send_callback(&sa->conn, skypeapi_buddy_pcb_get_info_2, \
		                            bui, "GET USER %s " ITEM "\n", who); \
	}

	SKYPEAPI_GET_INFO_ITEM(_("Contact Information"),
	                             _("Skype Name"), "HANDLE", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Full Name"), "FULLNAME", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(_("Personal Information"),
	                             _("Contacts"), "NROF_AUTHED_BUDDIES", 0, 0);
	//SKYPEAPI_GET_INFO_ITEM(NULL, , "TIMEZONE", 0, 0);
	//SKYPEAPI_GET_INFO_ITEM(NULL, , "BIRTHDAY", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Gender"), "SEX", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Language"), "LANGUAGE", 1, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Country"), "COUNTRY", 1, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Province"), "PROVINCE", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("City"), "CITY", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Home Phone"), "PHONE_HOME", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Office Phone"), "PHONE_OFFICE", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Mobile Phone"), "PHONE_MOBILE", 0, 0);
	SKYPEAPI_GET_INFO_ITEM(NULL, _("Homepage"), "HOMEPAGE", 0, 0);
	// make sure final item has 1 as last argument
	SKYPEAPI_GET_INFO_ITEM(NULL, _("About"), "ABOUT", 0, 1);

	// TODO: some other info, most importantly IS* and HAS*
#undef SKYPEAPI_GET_INFO_ITEM

	// BUG on timeout or something else that doesn't make the callback return,
	//     struct skypeapi_buddy_userdata_item and user_list are leaking.
}
static gboolean skypeapi_buddy_pcb_get_info_2(gpointer _gc, char *data, gpointer p)
{
	PurpleConnection *gc = (PurpleConnection *)_gc;
	struct skypeapi_purpleconn *sa = (struct skypeapi_purpleconn*) gc->proto_data;
	static PurpleNotifyUserInfo *user_info = NULL;
	struct skypeapi_buddy_userdata_item *bui = (struct skypeapi_buddy_userdata_item *)p; 
	char *cmd, *user, *item, *value;
	int i;

	// Get result from data
	cmd = strtok(data, skypeapi_conn_split_word);
	// TODO check command is indeed "USER" ?
	if (cmd) user = strtok(NULL, skypeapi_conn_split_word);
	if (user) item = strtok(NULL, skypeapi_conn_split_word);
		
	// strip number of words from beginning if requested
	// used for stripping ISO 639 prefix from language and country
	for (i=0; i<bui->skipwords && value && item; i++)
		value = strtok(NULL, skypeapi_conn_split_word); 

	if (bui->header)
		purple_notify_user_info_add_section_header(bui->user_info, bui->header);

	if (value) value = strtok(NULL, skypeapi_conn_split_eol);
	if (value) {
		// Add the pair; only "GENDER" needs some customization though ...
		if (g_ascii_strcasecmp(item, "GENDER")!=0 &&
		    g_ascii_strcasecmp(value, "UNKNOWN")!=0)
				purple_notify_user_info_add_pair(bui->user_info, bui->title, value);
	}

	if (bui->final) {
		// This is the last one, finish it and make it ready for next one
		purple_notify_userinfo(gc, user, bui->user_info, NULL, NULL);
		purple_notify_user_info_destroy(bui->user_info);
	}

	g_free(p);

	return FALSE;
}

