/**
 * @file conn.c
 *
 * Managing connection to Skype using one of the low-level connection methods
 *
 * Copyright (C) 2007, Willem van Engen <dev-pidgin@willem.engen.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdarg.h>
#include <string.h>
#include <glib/gprintf.h>

#include "conn.h"
#include "conn_debug.h"
#ifdef HAVE_GDK
#	include "conn_x.h"
#endif
#ifdef HAVE_DBUS
#	include "conn_dbus.h"
#endif

static void skypeapi_conn_cb_list_item_free(struct skypeapi_conn_cb *cb)
{
	g_free(cb->word);
	g_free(cb);
}

const char *skypeapi_conn_split_word = " \r\n\0";
const char *skypeapi_conn_split_eol  = "\r\n\0";
const char *skypeapi_conn_split_seq  = " ,\r\n\0";


// to be called by low-level connection when data received
void _skypeapi_conn_callback(void *data, char *cmd)
{
	struct skypeapi_conn *sc = (struct skypeapi_conn *)data;
	char *c, *newdata;
	GSList *cur = sc->cb_list;

	skypedbg_debug("<- %s\n", cmd);
	
	// TODO check for error 86 (also with msgid) and disconnect if so

	// don't crash on empty list :)
	if (!cur) return;

	c = strtok(cmd, skypeapi_conn_split_word);
	newdata = strtok(NULL, skypeapi_conn_split_eol);

	// call high-level callbacks if they match	
	do {
		struct skypeapi_conn_cb *cb = (struct skypeapi_conn_cb*)cur->data;

		if (!strcmp(c, cb->word) && cb->func) {
			// We have a match, invoke callback. Remember that callback may prepend
			// items to the current callback list, sc->cb_list.
			if ( !(cb->func(sc->cb_data_common, newdata, cb->userptr)) ) {
				// remove from list if requested
				// TODO can improve performance by remembering previous item and
				//      removing this item from that list, less traversing.
				sc->cb_list = g_slist_remove(sc->cb_list, (gpointer)cb);
				skypeapi_conn_cb_list_item_free(cb);
			}
		}		
	} while( (cur=g_slist_next(cur)) );	
}

gboolean skypeapi_conn_open(struct skypeapi_conn *sc, skypeapi_conn_type type, gpointer cb_data_common)
{
	int i;

	// set low-level functions
	switch(type) {
	case SKYPEAPI_CONN_TYPE_AUTO:
		// try the available types one by one
		for (i=SKYPEAPI_CONN_TYPE_AUTO+1; i<SKYPEAPI_CONN_TYPE_LAST; i++) {
			if (skypeapi_conn_open(sc, i, cb_data_common))
				return 1;
		}
		// and fail if none worked
		skypedbg_error("tried all connection types but none worked\n", type);
		return FALSE;

#ifdef HAVE_GDK
	case SKYPEAPI_CONN_TYPE_X:
		sc->_open = skypeapi_conn_x_open;
		sc->_close = skypeapi_conn_x_close;
		sc->_send_raw = skypeapi_conn_x_send_raw;
		break;
#endif
#ifdef HAVE_DBUS
	case SKYPEAPI_CONN_TYPE_DBUS_SESSION:
		sc->_open = skypeapi_conn_dbus_open_session;
		sc->_close = skypeapi_conn_dbus_close;
		sc->_send_raw = skypeapi_conn_dbus_send_raw;
		break;
	case SKYPEAPI_CONN_TYPE_DBUS_SYSTEM:
		sc->_open = skypeapi_conn_dbus_open_system;
		sc->_close = skypeapi_conn_dbus_close;
		sc->_send_raw = skypeapi_conn_dbus_send_raw;
		break;
#endif
	default:
		// not supported
		skypedbg_error("unsupported connection type %d requested\n", type);
		return FALSE;
	}
	// low-level open
	if (!sc->_open || !(sc->_conn=sc->_open(sc)) )
		return 0;
		
	// setup callbacks
	sc->cb_data_common = cb_data_common;
	sc->cb_list = NULL;

	return TRUE;
}

void skypeapi_conn_close(struct skypeapi_conn *sc)
{
	// might happen on _very_ early logout
	if (!sc) return;

	if (sc->_close) sc->_close(sc->_conn);
	
	g_slist_foreach(sc->cb_list, (GFunc)skypeapi_conn_cb_list_item_free, NULL);
	g_slist_free(sc->cb_list);
}


int skypeapi_conn_send(struct skypeapi_conn *sc, const char *fmt, ...)
{
	va_list args;
	int len;

	va_start(args, fmt);
	len = skypeapi_conn_send_v(sc, fmt, args);
	va_end(args);

	return len;
}

int skypeapi_conn_send_v(struct skypeapi_conn *sc, const char *fmt, va_list args)
{
	char *buf = NULL;
	int len;

	len = g_vasprintf(&buf, fmt, args);
	
	if (!len || !buf)
		return 0;

	skypedbg_debug("-> %s", buf); 
		
	if (sc->_send_raw && !sc->_send_raw(sc->_conn, buf, len)) {
		g_free(buf);
		return 0;
	}
		
	g_free(buf);
	return len;
}


int skypeapi_conn_send_callback(struct skypeapi_conn *sc, skypeapi_conn_callback_t cbfun,
                                gpointer userptr, const char *fmt, ...)
{
	va_list args;
	int len;

	va_start(args, fmt);
	len = skypeapi_conn_send_callback_v(sc, cbfun, userptr, fmt, args);
	va_end(args);

	return len;
}

int skypeapi_conn_send_callback_v(struct skypeapi_conn *sc, skypeapi_conn_callback_t cbfun,
                                  gpointer userptr, const char *fmt, va_list args)
{
	char *buf, *id;
	int len;

	// BUG #msgs exceeding 0xffffff might trigger old callbacks in list
	if ( !(id=g_strdup_printf("#%06x", sc->cb_index)) ) {
		return 0;
	}
	if ( !(buf=g_strdup_printf("%s %s", id, fmt)) ) {
		g_free(id);
		return 0;
	}
	if ( !skypeapi_conn_add_callback(sc, id, cbfun, userptr) ) {
		g_free(buf);
		g_free(id);
		return 0;
	}

	sc->cb_index++;
	len = skypeapi_conn_send_v(sc, buf, args);
	
	g_free(buf);
	g_free(id);
	return len;
}

gboolean skypeapi_conn_add_callback(struct skypeapi_conn *sc, const char *cmd, skypeapi_conn_callback_t cbfun, gpointer userptr)
{
	struct skypeapi_conn_cb *cb;

	cb=g_new0(struct skypeapi_conn_cb, 1);
	g_return_val_if_fail(cb!=NULL, FALSE);

	cb->word = g_strdup(cmd);
	g_return_val_if_fail(cb->word!=NULL, (g_free(cb), FALSE));
	cb->func = cbfun;
	cb->userptr = userptr;

	sc->cb_list = g_slist_prepend(sc->cb_list, cb);
	
	return TRUE;
}
