/**
 * @file conn.h
 *
 * Managing connection to Skype using one of the low-level connection methods
 *
 * Copyright (C) 2007, Willem van Engen <dev-pidgin@willem.engen.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef _PURPLE_SKYPEAPI_CONN_H_
#define _PURPLE_SKYPEAPI_CONN_H_

#include <stdarg.h>
#include <glib.h>
#include "config.h"

// callback for higher level layers
typedef gboolean (*skypeapi_conn_callback_t)(gpointer data_common, char *cmd, gpointer data);

// internal: low-level function callbacks, hands off if you're higher level
typedef void *(*_skypeapi_conn_open)(void *cb_data);
typedef void (*_skypeapi_conn_close)(void *sc);
typedef char (*_skypeapi_conn_send_raw)(void *sc, const char *buf, unsigned int len); 
void _sykeapi_conn_callback(void *data, char *cmd);

typedef enum skypeapi_conn_type_struct {
	SKYPEAPI_CONN_TYPE_AUTO = 0, // must be the first!
#ifdef HAVE_GDK
	SKYPEAPI_CONN_TYPE_X = 1,
#endif
#ifdef HAVE_DBUS
	SKYPEAPI_CONN_TYPE_DBUS_SESSION = 2,
	SKYPEAPI_CONN_TYPE_DBUS_SYSTEM = 3,
#endif
	SKYPEAPI_CONN_TYPE_LAST
} skypeapi_conn_type;

// internal: List of callbacks
struct skypeapi_conn_cb {
	char *word;						//< word to trigger callback on
	skypeapi_conn_callback_t func;	//< callback function
	void *userptr;					//< user pointer supplied to callback
};

// internal: Connection data
struct skypeapi_conn {
	// current callback token index for automated token generation
	int cb_index;
	// list of callbacks + common data
	GSList *cb_list;
	gpointer cb_data_common;
	
	// function pointers for low-level connection
	_skypeapi_conn_open _open;
	_skypeapi_conn_close _close;
	_skypeapi_conn_send_raw _send_raw;

	// low-level connection struct
	// this is a pointer to avoid includers having to include all connections
	void *_conn;
};

gboolean skypeapi_conn_open(struct skypeapi_conn *sc, skypeapi_conn_type type, gpointer cb_data_common);
void skypeapi_conn_close(struct skypeapi_conn *sc);
int skypeapi_conn_send(struct skypeapi_conn *sc, const char *fmt, ...);
int skypeapi_conn_send_v(struct skypeapi_conn *sc, const char *fmt, va_list args);
int skypeapi_conn_send_callback(struct skypeapi_conn *sc, skypeapi_conn_callback_t cbfun, gpointer userptr, const char *fmt, ...);
int skypeapi_conn_send_callback_v(struct skypeapi_conn *sc, skypeapi_conn_callback_t cbfun, gpointer userptr, const char *fmt, va_list args);
gboolean skypeapi_conn_add_callback(struct skypeapi_conn *sc, const char *cmd, skypeapi_conn_callback_t cbfun, gpointer userptr);

// useful for strtok
extern const char *skypeapi_conn_split_word;	// words
extern const char *skypeapi_conn_split_eol;		// 'till end of line
extern const char *skypeapi_conn_split_seq;		// items in a sequence

#endif /* _PURPLE_SKYPEAPI_CONN_H_ */
