/**
 * @file conn_x.c
 *
 * Low-level connection to Skype using X-messages.
 *
 * Copyright (C) 2007, Willem van Engen <dev-pidgin@willem.engen.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdlib.h>
#include <string.h>
#include <X11/Xatom.h> // for XA_WINDOW
#include <gdk/gdkx.h>

#include "conn_x.h"
#include "conn_debug.h"

struct skypeapi_conn_x { 	
	GdkNativeWindow win_skype;
	GdkWindow *win_self;

	GdkAtom atm_begin;
	GdkAtom atm_cont;

	void *cb_data;
	
	char rcvbuf[256]; // TODO: what length to use
	unsigned int len;
};

// list of currently active skypeapi_conn_x structs; we can only
// have one message callback and cannot remove it :(
static char first_run = 1;
static struct skypeapi_conn_x *_skypeapi_conn_x_current = NULL;

static char _skypeapi_conn_x_findskype(struct skypeapi_conn_x *sc)
{
	int status;
	Atom atm_skype = XInternAtom(GDK_DISPLAY(), "_SKYPE_INSTANCE", TRUE);
	Atom atm_type;
	int format;
	unsigned long len, len_after;
	guchar *prop_skype;

	/*
	 * GDK manual suggests to use XGetWindowProperty directly
	 * 
	GdkAtom atm_skype = gdk_atom_intern("_SKYPE_INSTANCE", TRUE);
	GdkAtom atm_type;
	status = gdk_property_get(gdk_get_default_root_window(),
	                 atm_skype, GDK_NONE, 0, 1, FALSE,
	                 &atm_type, &format, &len, &prop_skype);
	*/
	status = XGetWindowProperty(GDK_DISPLAY(), GDK_ROOT_WINDOW(),
	                            atm_skype, 0, 1, False, XA_WINDOW,
	                            &atm_type, &format, &len, &len_after, &prop_skype);

	if(status != Success || format != 32 || len != 1) {		
		skypedbg_error("Skype window not found, status=0x%x, format=0x%x, len=0x%x\n",
		               status, format, len);
		return 0;
	}

	sc->win_skype = (GdkNativeWindow)* (const unsigned long *) prop_skype & 0xffffffff;
	skypedbg_info("Found Skype window\n");
	return 1;
}

static GdkFilterReturn _skypeapi_conn_x_data_callback(GdkXEvent *_xev, GdkEvent *ev, gpointer data)
{
	struct skypeapi_conn_x *sc = _skypeapi_conn_x_current;
	XEvent *xev = (XEvent*)_xev;
	unsigned int i;
	
	g_return_val_if_fail(sc != NULL, GDK_FILTER_CONTINUE);

	g_return_val_if_fail(xev->xclient.format == 8, GDK_FILTER_CONTINUE);
	g_return_val_if_fail(xev->xclient.message_type == gdk_x11_atom_to_xatom(sc->atm_begin) ||
	                     xev->xclient.message_type == gdk_x11_atom_to_xatom(sc->atm_cont),
	                     GDK_FILTER_CONTINUE);

	// We should have seen the end message already and have reset len there
	if (xev->xclient.message_type == gdk_x11_atom_to_xatom(sc->atm_begin) && sc->len!=0) {
		skypedbg_error("incomplete packet of %d bytes discarded: %s\n",
			sc->len, sc->rcvbuf); 
		sc->len=0;
	}

	// Copy data to the buffer.
	for (i=0; i<20 && xev->xclient.data.b[i]!='\0'; i++) {
		if (sc->len < (sizeof(sc->rcvbuf)-1)) {
			sc->rcvbuf[sc->len++] = xev->xclient.data.b[i];
		}
	}
	sc->rcvbuf[sc->len]=0;

	// Last fragment is always <20 characters, so if we get one, the
	// middle layer is notified that data is ready
	if (i<20) {
		sc->len=0;
		// to upper layer
		_skypeapi_conn_callback(sc->cb_data, sc->rcvbuf);
	}

	// maybe others need it too
	return GDK_FILTER_CONTINUE;
}

/*
 * Public functions
 */
char skypeapi_conn_x_send_raw(void *_sc, const char *buf, unsigned int len)
{
	struct skypeapi_conn_x *sc = (struct skypeapi_conn_x *)_sc;
	unsigned int pos = 0;
	unsigned int i;
	XEvent xev;
	
	// begin message
	/*
	 * I wanted to use gdk's message send mechanism, but skype starts to
	 * use 100% cpu after it has received such a message. So I reverted
	 * to using Xlib calls.
	 *	 
	GdkEvent *ev;
	ev = gdk_event_new(GDK_CLIENT_EVENT);
	((GdkEventClient*)ev)->type = GDK_CLIENT_EVENT;
	((GdkEventClient*)ev)->window = sc->win_self;
	((GdkEventClient*)ev)->message_type = sc->atm_begin;
	((GdkEventClient*)ev)->data_format = 8;
	
	do
	{
		for( i=0; i < 20 && i+pos <= len; i++ )
			((GdkEventClient*)ev)->data.b[i] = buf[i+pos];
		gdk_event_send_client_message(ev, sc->win_skype);
		
		((GdkEventClient*)ev)->message_type = sc->atm_cont;
		pos += i;
	} while( pos <= len );

	*/ 
	memset(&xev, 0, sizeof(xev));
	xev.xclient.type = ClientMessage;
	xev.xclient.window = GDK_WINDOW_XID(sc->win_self);
	xev.xclient.message_type = gdk_x11_atom_to_xatom(sc->atm_begin);
	xev.xclient.format = 8;

	gdk_error_trap_push();

	// send buffer as messages in batches of 20 chars
	do
	{
		for( i=0; i < 20 && i+pos <= len; i++ )
			xev.xclient.data.b[i] = buf[i+pos];
			//((GdkEventClient*)ev)->data.b[i] = buf[i+pos];
		//gdk_event_send_client_message(ev, sc->win_skype);
		XSendEvent(GDK_DISPLAY(), sc->win_skype, False, 0, &xev);
		
		//((GdkEventClient*)ev)->message_type = sc->atm_cont;
		xev.xclient.message_type = gdk_x11_atom_to_xatom(sc->atm_cont);
		pos += i;
	} while( pos <= len );

	gdk_flush();

	if (gdk_error_trap_pop()) {
		skypedbg_warning("Failed to send command to Skype: %s\n", buf); // TODO should check NULL termination!
		return 0;
	}	
	
	return 1;	
}

void *skypeapi_conn_x_open(void *cb_data)
{
	struct skypeapi_conn_x *sc;
	GdkWindowAttr attr;
	
	// We allow only one instance. It'd be hard anyway to
	// distinguish to different skype instances, if at all possible.
	// It makes the workaround for the missing
	// gdk_remove_client_message_filter() somewhat easier.
	// TODO: allow more DISPLAYs and subsequently multiple instances of plugin
	if (_skypeapi_conn_x_current) {
		skypedbg_error("Cannot run than more instance of the SkypeAPI plugin\n");
		return NULL;
	}

	// allocate new me
	if ( !(sc=g_new0(struct skypeapi_conn_x,1)) )
		return NULL;

	// Create window for sending messages
	attr.x=-1; attr.y=-1; attr.width=0; attr.height=0;
	attr.wclass = GDK_INPUT_ONLY;
	sc->win_self = gdk_window_new(NULL, &attr, GDK_WA_X|GDK_WA_Y);
	if (!sc->win_self) {
		skypedbg_error("Couldn't create message window\n");
		g_free(sc);
		return NULL;
	} 
	
	if (!_skypeapi_conn_x_findskype(sc)) {
		skypedbg_error("Couldn't find Skype program\n");
		g_free(sc);
		return NULL;
	}

	// must not add g_x11_xatom_to_atom or gdk_add_client_message_filter won't work
	sc->atm_begin = gdk_atom_intern("SKYPECONTROLAPI_MESSAGE_BEGIN", FALSE);
	sc->atm_cont  = gdk_atom_intern("SKYPECONTROLAPI_MESSAGE", FALSE);
	
	_skypeapi_conn_x_current = sc;
	sc->cb_data = cb_data;
	sc->len = 0;
	
	if (first_run) {
		gdk_add_client_message_filter(sc->atm_begin, _skypeapi_conn_x_data_callback, NULL); 
		gdk_add_client_message_filter(sc->atm_cont,  _skypeapi_conn_x_data_callback, NULL); 
		skypedbg_info("Skype message hooks activated\n");
		first_run = 0;
	}
	
	return (void*)sc;
}

void skypeapi_conn_x_close(void *_sc)
{
	struct skypeapi_conn_x *sc = (struct skypeapi_conn_x *)_sc;
	// callback musn't update this anymore when freed
	_skypeapi_conn_x_current = NULL;

	if (!sc) return;

	// destroy window
	gdk_window_destroy(sc->win_self);
	// and get rid of myself
	g_free(sc);
}

