/**
 * @file skypeapi.c
 *
 * purple protocol plugin for the Skype API
 *
 * Copyright (C) 2007, Willem van Engen <dev-pidgin@willem.engen.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <string.h>
#include <glib/gprintf.h>

#include <debug.h>
#include <accountopt.h>
#include <notify.h>
#include <prpl.h>
#include <plugin.h>
#include <util.h>
#include <version.h>

#include "config.h"
#include "skypeapi.h"
#include "buddy.h"

// Name of program announced to Skype, displayed in authorization prompt
#define SKYPEAPI_API_NAME "Libpurple-SkypeAPI"
// ID for libpurple
#define SKYPEAPI_PURPLE_ID "prpl-skypeapi"


// callbacks need to be declared
static gboolean skypeapi_login_2(gpointer _gc, char *data, gpointer p);
static gboolean skypeapi_login_2_tryagain(gpointer _gc);
static gboolean skypeapi_login_3(gpointer _gc, char *data, gpointer p);

gboolean skypeapi_api_check_ok(const char *data)
{
	// TODO make sure no data is present after OK, ever
	return strcmp(data, "OK")==0;
}

// returns error number from a skype reply, or zero if none
// optionally returns message string if there is one
int skypeapi_api_check_error(const char *data, char **errmsg)
{
	char *pos, *myerrmsg=NULL;
	int err = 0;

	// TODO make sure we have nothing like "ERRORFOO"
	//      ok, just make this function nicer.
	if (g_str_has_prefix(data, "ERROR")) {
		err = -1; // still an error if parsing of remainder fails
		pos = strchr(data, ' ');
		if (pos) {
			err = atoi(&pos[1]);
			pos = strchr(&pos[1], ' ');
			if (pos && pos[1]) {
				myerrmsg = &pos[1];
			}
		}
	}

	if (errmsg) *errmsg=pos;
	return err;
}


static GList *skypeapi_status_types(PurpleAccount *account)
{
	PurpleStatusType *type;
	GList *types = NULL;
	
	purple_debug_info("skypeapi", "Enumerating account status types\n");
	
	type = purple_status_type_new(PURPLE_STATUS_AVAILABLE, "ONLINE", NULL, TRUE);
	types = g_list_append(types, type);

	type = purple_status_type_new(PURPLE_STATUS_AVAILABLE, "SKYPEME", _("Skype Me"), TRUE);
	types = g_list_append(types, type);

	type = purple_status_type_new(PURPLE_STATUS_AWAY, "AWAY", NULL, TRUE);
	types = g_list_append(types, type);

	type = purple_status_type_new(PURPLE_STATUS_UNAVAILABLE, "NA", _("Not Available"), TRUE);
	types = g_list_append(types, type);

	type = purple_status_type_new(PURPLE_STATUS_UNAVAILABLE, "DND", _("Do Not Disturb"), TRUE);
	types = g_list_append(types, type);

	type = purple_status_type_new(PURPLE_STATUS_INVISIBLE, "INVISIBLE", NULL, TRUE);
	types = g_list_append(types, type);

	type = purple_status_type_new(PURPLE_STATUS_OFFLINE, "OFFLINE", NULL, TRUE);
	types = g_list_append(types, type);

	return types;
}

// number of stages in login, for purple_connection_update_progess()
#define SKYPEAPI_LOGIN_STAGES 3

static void skypeapi_login(PurpleAccount *acc) {
	PurpleConnection *gc = purple_account_get_connection(acc);
	struct skypeapi_purpleconn *sa;
	int conn_method;

	sa = g_new0(struct skypeapi_purpleconn, 1);
	g_return_if_fail( sa != NULL );
	sa->account = acc;
	gc->proto_data = sa;
	sa->handle_msg = purple_account_get_bool(acc, "handle_msg", FALSE);
	conn_method = atoi(purple_account_get_string(acc, "conn_method", "0"));
	
	purple_connection_update_progress(gc, _("Connecting"), 0, SKYPEAPI_LOGIN_STAGES);

	if ( !skypeapi_conn_open(&sa->conn, conn_method, gc) ) {
		purple_connection_error(gc,
			_("Could not connect to Skype program, make sure you have started Skype already."));
		return;	
	}

	purple_connection_update_progress(gc, _("Negotiating"), 1, SKYPEAPI_LOGIN_STAGES);
	
	skypeapi_conn_send_callback(&sa->conn, skypeapi_login_2, NULL, "NAME " SKYPEAPI_API_NAME "\n");
	
	return;
}
static gboolean skypeapi_login_2(gpointer _gc, char *data, gpointer p)
{
	PurpleConnection *gc = (PurpleConnection *)_gc;
	struct skypeapi_purpleconn *sa = gc->proto_data;
	char *c, *errstr;
	int err;
	
	if ( (err=skypeapi_api_check_error(data,&errstr)) ) {
		if (err==68) {
			// access denied
			purple_connection_error(gc,
				_("Access denied, please allow the program called \"Libpurple-SkypeAPI\" to connect to Skype.\n"
				  "To achieve this, open the Skype preferences, select \"Public API\" and remove the aforementioned "
				  "program from the list of blocked programs. Then reconnect to Skype from here."));
		} else {
			char *errmsg = g_strdup_printf(("Connection error: %s\n"), errstr);
			purple_connection_error(gc, errmsg);
			free(errmsg);
		}
		return FALSE;
	}

	if (skypeapi_api_check_ok(data)) {
		skypeapi_conn_send_callback(&sa->conn, skypeapi_login_3, NULL, "PROTOCOL 7\n");
	} else {
		c = strtok(data, skypeapi_conn_split_word);
		if (g_ascii_strcasecmp(c, "CONNSTATUS")!=0)
			purple_debug_warning("skypeapi", "Unexpected reply to NAME: %s\n", data);
		// Try again after some time
		// TODO limit number of tries?
		g_timeout_add_seconds(10, skypeapi_login_2_tryagain, gc);
	}
	return FALSE;
}
static gboolean skypeapi_login_2_tryagain(gpointer _gc)
{
	PurpleConnection *gc = (PurpleConnection *)_gc;
	struct skypeapi_purpleconn *sa = gc->proto_data;
	skypeapi_conn_send_callback(&sa->conn, skypeapi_login_2, NULL, "NAME " SKYPEAPI_API_NAME "\n");
	return FALSE; // don't call again
}
static gboolean skypeapi_login_3(gpointer _gc, char *data, gpointer p)
{
	PurpleConnection *gc = (PurpleConnection *)_gc;
	PurpleAccount *acc = purple_connection_get_account(gc);
	struct skypeapi_purpleconn *sa = gc->proto_data;
	char *c, *protvers;
	
	c = strtok(data, skypeapi_conn_split_word);
	protvers = strtok(NULL, skypeapi_conn_split_word);
	sa->protoversion = atoi(protvers);
	purple_debug_info("skypeapi", "got protocol version %d\n", sa->protoversion);
	
	purple_connection_update_progress(gc, _("Synchronizing"), 2, SKYPEAPI_LOGIN_STAGES);
	
	// Initialize all components
	skypeapi_buddy_init(gc);

	purple_connection_set_state(gc, PURPLE_CONNECTED);

	return FALSE;
}

static void skypeapi_close(PurpleConnection *gc)
{
	struct skypeapi_purpleconn *sa = gc->proto_data;

	if (sa == NULL)
		return;
	
	// TODO: maybe send close message

	skypeapi_buddy_cleanup(gc);
	
	skypeapi_conn_close(&sa->conn);
	
	g_free(gc->proto_data);
	gc->proto_data = NULL;
}


static void skypeapi_set_status(PurpleAccount *account, PurpleStatus *status)
{
	PurpleConnection *gc = purple_account_get_connection(account);
	const char *status_id = purple_status_get_id(status);
	struct skypeapi_purpleconn *sa;

	purple_debug_info("skypeapi", "set status to %s\n", status_id);

	g_return_if_fail(gc != NULL);
	sa = gc->proto_data;

	//if ( !purple_status_is_active(status) )
	//	return;
		
	skypeapi_conn_send(&sa->conn, "SET USERSTATUS %s\n", status_id);
	
	return;
}

static const char* skypeapi_list_icon(PurpleAccount *a, PurpleBuddy *b)
{
	return "skype";
}

static PurplePluginProtocolInfo prpl_info =
{
	OPT_PROTO_NO_PASSWORD | OPT_PROTO_CHAT_TOPIC,
	NULL,						/* user_splits */
	NULL,						/* protocol_options */
	NO_BUDDY_ICONS,				/* icon_spec */
	skypeapi_list_icon,			/* list_icon */
	NULL,						/* list_emblems */
	NULL,						/* status_text */
	NULL,						/* tooltip_text */
	skypeapi_status_types,		/* away_states */
	NULL,	/* blist_node_menu */
	NULL,						/* chat_info */
	NULL,						/* chat_info_defaults */
	skypeapi_login,				/* login */
	skypeapi_close,				/* close */
	NULL,						/* send_im */
	NULL,						/* set_info */
	NULL,						/* send_typing */
	skypeapi_buddy_pcb_get_info,/* get_info */
	skypeapi_set_status,		/* set_status */
	NULL,						/* set_idle */
	NULL,						/* change_passwd */
	skypeapi_buddy_pcb_add,		/* add_buddy */
	NULL,						/* add_buddies */
	skypeapi_buddy_pcb_rem,		/* remove_buddy */
	NULL,						/* remove_buddies */
	NULL,						/* add_permit */
	NULL,						/* add_deny */
	NULL,						/* rem_permit */
	NULL,						/* rem_deny */
	NULL,						/* set_permit_deny */
	NULL,		/* join_chat */
	NULL,						/* reject_chat */
	NULL,		/* get_chat_name */
	NULL,		/* chat_invite */
	NULL,		/* chat_leave */
	NULL,						/* chat_whisper */
	NULL,		/* chat_send */
	NULL,		/* keepalive */
	NULL,						/* register_user */
	NULL,						/* get_cb_info */
	NULL,						/* get_cb_away */
	NULL,						/* alias_buddy */
	NULL,						/* group_buddy */
	NULL,						/* rename_group */
	NULL,						/* buddy_free */
	NULL,						/* convo_closed */
	NULL,		/* normalize */
	NULL,						/* set_buddy_icon */
	NULL,						/* remove_group */
	NULL,						/* get_cb_real_name */
	NULL,						/* set_chat_topic */
	NULL,						/* find_blist_chat */
	NULL,						/* roomlist_get_list */
	NULL,						/* roomlist_cancel */
	NULL,						/* roomlist_expand_category */
	NULL,						/* can_receive_file */
	NULL,						/* send_file */
	NULL,						/* new_xfer */
	NULL,						/* offline_message */
	NULL,						/* whiteboard_prpl_ops */
	NULL,						/* send_raw */
	NULL,						/* roomlist_room_serialize */

	/* padding */
	NULL,
	NULL,
	NULL,
	NULL
};

static PurplePluginInfo info =
{
	PURPLE_PLUGIN_MAGIC,
	PURPLE_MAJOR_VERSION,
	PURPLE_MINOR_VERSION,
	PURPLE_PLUGIN_PROTOCOL,                           /**< type           */
	NULL,                                             /**< ui_requirement */
	0,                                                /**< flags          */
	NULL,                                             /**< dependencies   */
	PURPLE_PRIORITY_DEFAULT,                          /**< priority       */

	SKYPEAPI_PURPLE_ID,                               /**< id             */
	"SkypeAPI",                                       /**< name           */
	PACKAGE_VERSION,                                  /**< version        */
	N_("Skype API Protocol Plugin"),                  /**  summary        */
	N_("Skype Protocol Plugin that uses the Skype API to connect to a running instance of Skype"),    /**  description    */
	N_("Willem van Engen <dev-pidgin@willem.engen.nl>"),/**< author         */
	"http://willem.engen.nl/projects/libpurple-skypeapi/",  /**< homepage       */

	NULL,		                                      /**< load           */
	NULL,                                             /**< unload         */
	NULL,                                             /**< destroy        */

	NULL,                                             /**< ui_info        */
	&prpl_info,                                       /**< extra_info     */
	NULL,                                             /**< prefs_info     */
	NULL,

	/* padding */
	NULL,
	NULL,
	NULL,
	NULL
};

static void _init_plugin(PurplePlugin *plugin)
{
	PurpleAccountUserSplit *split;
	PurpleAccountOption *option;

#if 0 // Not yet -- chat temporarily not available
	// Whether to handle incoming chat messages or not
	option = purple_account_option_bool_new(
    	_("Handle chat messages"),       /* text shown to user */
    	"handle_msg",                    /* pref name */
    	FALSE);                          /* default value */
	prpl_info.protocol_options = g_list_append(prpl_info.protocol_options, option);
#endif

	// Connection method
	option = purple_account_option_list_new(
		_("Connection method"),
		"conn_method",
		NULL);
	purple_account_option_add_list_item(option, _("Automatic"), "0");
#ifdef HAVE_GDK
	purple_account_option_add_list_item(option, _("X messages"), "1");
#endif
#ifdef HAVE_DBUS
	purple_account_option_add_list_item(option, _("D-Bus (session)"), "2");
	purple_account_option_add_list_item(option, _("D-Bus (system)"), "3");
#endif
	prpl_info.protocol_options = g_list_append(prpl_info.protocol_options, option);

	purple_prefs_remove("/plugins/prpl/skypeapi");

	//skypeapi_register_commands();
}

PURPLE_INIT_PLUGIN(skypeapi, _init_plugin, info);
